from ..core import RomanDataModel


class ReferenceFileModel(RomanDataModel):
    """
    Data model for reference files.
    """
    schema_url = "http://stsci.edu/schemas/roman_datamodel/reference_files/referencefile.schema"

    def __init__(self, init=None, **kwargs):
        super().__init__(init=init, **kwargs)



"""Base classes for Roman tests"""
from glob import glob as _sys_glob
import os
from os import path as op
from pathlib import Path
import sys
import pytest
import requests

from ci_watson.artifactory_helpers import (
    BigdataError,
    check_url,
    get_bigdata,
    get_bigdata_root,
)

# from jwst.associations import load_asn

__all__ = [
    'BaseRomanTest',
]

# Define location of default Artifactory API key, for Jenkins use only
ARTIFACTORY_API_KEY_FILE = '/eng/ssb2/keys/svc_rodata.key'


@pytest.mark.usefixtures('_jail')
@pytest.mark.bigdata
class BaseRomanTest:
    '''
    Base test class from which to derive Roman regression tests
    '''
    rtol = 0.00001
    atol = 0

    input_loc = ''  # root directory for 'input' files
    ref_loc = []  # root path for 'truth' files: ['test1','truth'] or ['test3']

    ignore_table_keywords = []
    ignore_fields = []
    ignore_hdus = ['ASDF']
    ignore_keywords = ['DATE', 'CAL_VER', 'CAL_VCS', 'CRDS_VER',
                       'CRDS_CTX', 'FILENAME']

    @pytest.fixture(autouse=True)
    def config_env(self, pytestconfig, envopt):
        self.env = pytestconfig.getoption('env')

    @pytest.fixture(autouse=True)
    def config_access(self, pytestconfig):
        self.inputs_root = pytestconfig.getini('inputs_root')[0]
        self.results_root = pytestconfig.getini('results_root')[0]

    @property
    def repo_path(self):
        return [self.inputs_root, self.env, self.input_loc]

    def get_data(self, *pathargs, docopy=True):
        """
        Download `filename` into working directory using
        `artifactory_helpers/get_bigdata()`.
        This will then return the full path to the local copy of the file.
        """
        local_file = get_bigdata(*self.repo_path, *pathargs, docopy=docopy)
        return local_file

    def data_glob(self, *pathargs, glob='*'):
        """Retrieve file list matching glob

        Parameters
        ----------
        pathargs: (str[, ...])
            Path components

        glob: str
            The file name match criterion

        Returns
        -------
        file_paths: [str[, ...]]
            File paths that match the glob criterion.
            Note that the TEST_BIGDATA and `repo_path`
            roots are removed so these results can be fed
            back into `get_data()`
        """

        # Get full path and proceed depending on whether
        # is a local path or URL.
        root = get_bigdata_root()
        if op.exists(root):
            path = op.join(root, *self.repo_path)
            root_len = len(path) + 1
            path = op.join(path, *pathargs)
            file_paths = _data_glob_local(path, glob)
        elif check_url(root):
            root_len = len(op.join(*self.repo_path[1:])) + 1
            path = op.join(*self.repo_path, *pathargs)
            file_paths = _data_glob_url(path, glob, root=root)
        else:
            raise BigdataError('Path cannot be found: {}'.format(path))

        # Remove the root from the paths
        file_paths = [
            file_path[root_len:]
            for file_path in file_paths
        ]
        return file_paths


# def raw_from_asn(asn_file):
#     """
#     Return a list of all input files from a given association.
#
#     Parameters
#     ----------
#     asn_file : str
#         Filename for the ASN file.
#
#     Returns
#     -------
#     members : list of str
#         A list of all input files in the association
#
#     """
#
#     members = []
#     # with open(asn_file) as f:
#     #     asn = load_asn(f)
#
#     for product in asn['products']:
#         for member in product['members']:
#             members.append(member['expname'])
#
#     return members
#

def _data_glob_local(*glob_parts):
    """Perform a glob on the local path

    Parameters
    ----------
    glob_parts: (path-like,[...])
        List of components that will be built into a single path

    Returns
    -------
    file_paths: [str[, ...]]
        Full file paths that match the glob criterion
    """
    full_glob = Path().joinpath(*glob_parts)
    return _sys_glob(str(full_glob))


def _data_glob_url(*url_parts, root=None):
    """
    Parameters
    ----------
    url: (str[,...])
        List of components that will be used to create a URL path

    root: str
        The root server path to the Artifactory server.
        Normally retrieved from `get_bigdata_root`.

    Returns
    -------
    url_paths: [str[, ...]]
        Full URLS that match the glob criterion
    """
    # Fix root root-ed-ness
    if root.endswith('/'):
        root = root[:-1]

    # Access
    try:
        envkey = os.environ['API_KEY_FILE']
    except KeyError:
        envkey = ARTIFACTORY_API_KEY_FILE

    try:
        with open(envkey) as fp:
            headers = {'X-JFrog-Art-Api': fp.readline().strip()}
    except (PermissionError, FileNotFoundError):
        print("Warning: Anonymous Artifactory search requests are limited "
              "to 1000 results. Use an API key and define API_KEY_FILE "
              "environment variable to get full search results.",
              file=sys.stderr)
        headers = None

    search_url = '/'.join([root, 'api/search/pattern'])

    # Join and re-split the url so that every component is identified.
    url = '/'.join([root] + [idx for idx in url_parts])
    all_parts = url.split('/')

    # Pick out "romancal-pipeline", the repo name
    repo = all_parts[4]

    # Format the pattern
    pattern = repo + ':' + '/'.join(all_parts[5:])

    # Make the query
    params = {'pattern': pattern}
    with requests.get(search_url, params=params, headers=headers) as r:
        url_paths = r.json()['files']

    return url_paths



from datetime import datetime
import copy
import json
import os
from pathlib import Path

import getpass
import pytest
from ci_watson.artifactory_helpers import UPLOAD_SCHEMA
from astropy.table import Table
from numpy.testing import assert_allclose, assert_equal
#from astropy.io.fits import conf

from romancal.regtest.regtestdata import RegtestData
from romancal.regtest.sdp_pools_source import SDPPoolsSource


TODAYS_DATE = datetime.now().strftime("%Y-%m-%d")

# # Turn of FITS memmap for all regtests (affects FITSDiff)
# conf.use_memmap = False


@pytest.fixture(scope="session")
def artifactory_repos(pytestconfig):
    """Provides Artifactory inputs_root and results_root"""
    try:
        inputs_root = pytestconfig.getini('inputs_root')[0]
        results_root = pytestconfig.getini('results_root')[0]
    except IndexError:
        inputs_root = "romancal-pipeline"
        results_root = "romancal-pipeline-results"
    return inputs_root, results_root


@pytest.hookimpl(tryfirst=True, hookwrapper=True)
def pytest_runtest_makereport(item, call):
    """Create request.node.report_setup and request.node.report_call to be
    used in generate_artifactory_json() fixture.
    """
    # execute all other hooks to obtain the report object
    outcome = yield
    rep = outcome.get_result()

    # set a report attribute for each phase of a call, which can
    # be "setup", "call", "teardown"
    setattr(item, "report_" + rep.when, rep)


def postmortem(request, fixturename):
    """Retrieve a fixture object if a test failed, else return None
    """
    try:
        if request.node.report_setup.passed:
            try:
                if request.node.report_call.failed:
                    return request.node.funcargs.get(fixturename, None)
            # Handle case where `report_call` hasn't been generated yet
            # because the test hasn't finished running, as in the case of
            # a user interrupt
            except AttributeError:
                return None
    except AttributeError:
        return None


@pytest.fixture(scope='function', autouse=True)
def generate_artifactory_json(request, artifactory_repos):
    """Pytest fixture that leaves behind JSON upload and okify specfiles
    if the rtdata or rtdata_module fixtures are used in either a test or a
    module-scoped fixture that runs a pipeline and provides the results to a
    series of test.
    """
    inputs_root, results_root = artifactory_repos

    def artifactory_result_path():
        # Generate the Artifactory result path
        whoami = getpass.getuser() or 'nobody'
        user_tag = 'NOT_CI_{}'.format(whoami)
        build_tag = os.environ.get('BUILD_TAG', user_tag)
        build_matrix_suffix = os.environ.get('BUILD_MATRIX_SUFFIX', '0')
        subdir = '{}_{}_{}'.format(TODAYS_DATE, build_tag, build_matrix_suffix)
        testname = request.node.originalname or request.node.name

        return os.path.join(results_root, subdir, testname) + os.sep

    yield
    # Execute the following at test teardown
    upload_schema_pattern = []
    okify_schema_pattern = []

    rtdata = postmortem(request, 'rtdata') or postmortem(request, 'rtdata_module')
    if rtdata:
        try:
            # The _jail fixture from ci_watson sets tmp_path
            cwd = str(request.node.funcargs['tmp_path'])
        except KeyError:
            # The jail fixture (module-scoped) returns the path
            cwd = str(request.node.funcargs['jail'])
        rtdata.remote_results_path = artifactory_result_path()
        rtdata.test_name = request.node.name
        # Dump the failed test traceback into rtdata
        rtdata.traceback = str(request.node.report_call.longrepr)

        # Upload and allow okify of truth by rtdata.output, if the test did not
        # fail before producing rtdata.output
        if rtdata.output and os.path.exists(rtdata.output):
            # Write the rtdata class out as an ASDF file
            path_asdf = os.path.join(cwd, f"{request.node.name}_rtdata.asdf")
            rtdata.to_asdf(path_asdf)

            # Generate an OKify JSON file
            pattern = os.path.join(rtdata.remote_results_path,
                os.path.basename(rtdata.output))
            okify_schema_pattern.append(pattern)
            okify_schema = generate_upload_schema(okify_schema_pattern,
                f"{os.path.dirname(rtdata.truth_remote)}/")

            jsonfile = os.path.join(cwd, f"{request.node.name}_okify.json")
            with open(jsonfile, 'w') as fd:
                json.dump(okify_schema, fd, indent=2)

            # Generate an upload JSON file, including the OKify, asdf file
            upload_schema_pattern.append(rtdata.output)
            upload_schema_pattern.append(os.path.abspath(jsonfile))
            upload_schema_pattern.append(path_asdf)
            upload_schema = generate_upload_schema(upload_schema_pattern,
                rtdata.remote_results_path)

            jsonfile = os.path.join(cwd, f"{request.node.name}_results.json")
            with open(jsonfile, 'w') as fd:
                json.dump(upload_schema, fd, indent=2)


def generate_upload_schema(pattern, target, recursive=False):
    """
    Generate JSON schema for Artifactory upload specfile using JFROG.

    Docs can be found at
    https://www.jfrog.com/confluence/display/RTF/Using+File+Specs

    Parameters
    ----------
    pattern : str or list of strings
        Specifies the local file system path to test results which should be
        uploaded to Artifactory. You can specify multiple artifacts by using
        wildcards or a regular expression as designated by the regexp property.

    target : str
        Specifies the target path in Artifactory in the following format::

            [repository_name]/[repository_path]

    recursive : bool, optional
        Specify whether or not to identify files listed in sub-directories
        for uploading.  Default: `False`

    Returns
    -------
    upload_schema : dict
        Dictionary specifying the upload schema
    """
    recursive = repr(recursive).lower()

    if not isinstance(pattern, str):
        # Populate schema for this test's data
        upload_schema = {"files": []}

        for p in pattern:
            temp_schema = copy.deepcopy(UPLOAD_SCHEMA["files"][0])
            temp_schema.update({"pattern": p, "target": target,
                                "recursive": recursive})
            upload_schema["files"].append(temp_schema)
    else:
        # Populate schema for this test's data
        upload_schema = copy.deepcopy(UPLOAD_SCHEMA)
        upload_schema["files"][0].update({"pattern": pattern, "target": target,
                                          "recursive": recursive})
    return upload_schema


def _rtdata_fixture_implementation(artifactory_repos, envopt, request):
    """Provides the RemoteResource class"""
    inputs_root, results_root = artifactory_repos
    rtdata = RegtestData(env=envopt, inputs_root=inputs_root,
        results_root=results_root)

    yield rtdata


@pytest.fixture(scope='function')
def rtdata(artifactory_repos, envopt, request, _jail):
    yield from _rtdata_fixture_implementation(artifactory_repos, envopt, request)


@pytest.fixture(scope='module')
def rtdata_module(artifactory_repos, envopt, request, jail):
    yield from _rtdata_fixture_implementation(artifactory_repos, envopt, request)


@pytest.fixture
def ignore_asdf_paths():
    ignore_attr = ["meta.[date, filename]",
                   "asdf_library",
                   "history"]

    return {'ignore': ignore_attr}


@pytest.fixture
def diff_astropy_tables():
    """Compare astropy tables with tolerances for float columns."""

    def _diff_astropy_tables(result_path, truth_path, rtol=1e-5, atol=1e-7):
        __tracebackhide__ = True
        result = Table.read(result_path)
        truth = Table.read(truth_path)

        diffs = []

        try:
            assert result.colnames == truth.colnames
        except AssertionError as err:
            diffs.append(f"Column names (or order) do not match\n{err}")

        try:
            assert len(result) == len(truth)
        except AssertionError as err:
            diffs.append(f"Row count does not match\n{err}")

        # If either the columns or the row count is mismatched, then don't
        # bother checking the individual column values.
        if len(diffs) > 0:
            raise AssertionError("\n".join(diffs))

        # Disable meta comparison for now, until we're able to specify
        # individual entries for comparison
        #if result.meta != truth.meta:
        #    diffs.append("Metadata does not match")

        for col_name in truth.colnames:
            try:
                try:
                    assert result[col_name].dtype == truth[col_name].dtype
                except AssertionError as err:
                    diffs.append(f"Column '{col_name}' dtype does not match\n{err}")
                    continue

                if truth[col_name].dtype.kind == "f":
                    try:
                        assert_allclose(result[col_name], truth[col_name],
                            rtol=rtol, atol=atol)
                    except AssertionError as err:
                        diffs.append("\n----------------------------------\n"
                            + f"Column '{col_name}' values do not "
                            + f"match (within tolerances) \n{err}"
                        )
                else:
                    try:
                        assert_equal(result[col_name], truth[col_name])
                    except AssertionError as err:
                        diffs.append(f"Column '{col_name}' values do not match\n{err}")
            except AttributeError:
                # Ignore case where a column does not have a dtype, as in the case
                # of SkyCoord objects
                pass

        if len(diffs) != 0:
            raise AssertionError("\n".join(diffs))

        # No differences
        return True

    return _diff_astropy_tables

# Add option to specify a single pool name
def pytest_addoption(parser):
    parser.addoption(
        '--sdp-pool', metavar='sdp_pool', default=None,
        help='SDP test pool to run. Specify the name only, not extension or path'
    )
    parser.addoption(
        '--standard-pool', metavar='standard_pool', default=None,
        help='Standard test pool to run. Specify the name only, not extension or path'
    )


@pytest.fixture
def sdp_pool(request):
    """Retrieve a specific SDP pool to test"""
    return request.config.getoption('--sdp-pool')


@pytest.fixture
def standard_pool(request):
    """Retrieve a specific standard pool to test"""
    return request.config.getoption('--standard-pool')


def pytest_generate_tests(metafunc):
    """Prefetch and parametrize a set of test pools"""
    if 'pool_path' in metafunc.fixturenames:
        try:
            SDPPoolsSource.inputs_root = metafunc.config.getini('inputs_root')[0]
            SDPPoolsSource.results_root = metafunc.config.getini('results_root')[0]
            SDPPoolsSource.env = metafunc.config.getoption('env')
        except IndexError:
            SDPPoolsSource.inputs_root = "romancal-pipeline"
            SDPPoolsSource.results_root = "romancal-pipeline-results"
            SDPPoolsSource.env = "dev"

        pools = SDPPoolsSource()

        try:
            pool_paths = pools.pool_paths
        except Exception:
            pool_paths = []

        ids = [
            Path(pool_path).stem
            for pool_path in pool_paths
        ]

        metafunc.parametrize('pool_path', pool_paths, ids=ids)



"""
Unit tests for dark current correction
"""

import warnings
import pytest
import numpy as np
from numpy.testing import assert_allclose
from astropy.time import Time

from romancal.dark_current.dark_sub import (
    average_dark_frames,
    do_correction as darkcorr
    )
from romancal.datamodels import RampModel, DarkModel, dqflags


# Define frame_time and number of groups in the generated dark reffile
TFRAME = 10.0
NGROUPS_DARK = 10


def _params():
    """ Returns list of tuples, one for several combinations of ngroups,
        nframes, and nskip (which are some of the parameter which will later
        be retrieved from the MA table), generating parameters for
        test_frame_averaging.  Parameters are the following:

        (ma_tab, ngroups, nframes, groupgap, nrows, ncols)
    """
    # Dictionary of pseudo readout patterns
    ma_tab_infos = dict(
        RP1=dict(ngroups=20, nframes=8, nskip=0),
        RP2=dict(ngroups=32, nframes=2, nskip=0),
        RP3=dict(ngroups=12, nframes=6, nskip=0),
    )

    params = []
    ngroups = 3

    # WFI is 4096x4096, but we reduce the size to 20x20 for speed/memory
    nrows = 20
    ncols = 20
    for ma_tab_info, values in ma_tab_infos.items():
        params.append((ma_tab_info, ngroups, values['nframes'],
                       values['nskip'], nrows, ncols))

    return params


@pytest.mark.parametrize('ma_tab_info, ngroups, nframes, groupgap, nrows,'
                         'ncols', _params())
def test_frame_averaging(setup_nrc_cube, ma_tab_info, ngroups, nframes,
                         groupgap, nrows, ncols):

    '''Check that if nframes>1 or groupgap>0, then the pipeline reconstructs
       the dark reference file to match the frame averaging and groupgap
       settings of the exposure.'''

    # Create data and dark model
    data, dark = setup_nrc_cube(ma_tab_info, ngroups, nframes, groupgap,
                                nrows, ncols)

    # Add ramp values to dark model data array
    dark.data[:, 10, 10] = np.arange(0, NGROUPS_DARK)
    dark.err[:, 10, 10] = np.arange(10, NGROUPS_DARK + 10)

    # Run the pipeline's averaging function
    avg_dark = average_dark_frames(dark, ngroups, nframes, groupgap)

    # Group input groups into collections of frames which will be averaged
    total_frames = (nframes * ngroups) + (groupgap * (ngroups-1))

    # Get starting/ending indexes of the input groups to be averaged
    gstrt_ind = np.arange(0, total_frames, nframes + groupgap)
    gend_ind = gstrt_ind + nframes

    # Prepare arrays to hold results of averaging
    manual_avg = np.zeros((ngroups), dtype=np.float32)
    manual_errs = np.zeros((ngroups), dtype=np.float32)

    # Manually average the input data to compare with pipeline output
    for newgp, gstart, gend in zip(range(ngroups), gstrt_ind, gend_ind):

        # Average the data frames
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=RuntimeWarning)
            newframe = np.mean(dark.data[gstart:gend, 10, 10])

            manual_avg[newgp] = newframe

        # ERR arrays will be quadratic sum of error values
        err_array = np.sum(dark.err[gstart:gend, 10, 10]**2)
        manual_errs[newgp] = np.sqrt(err_array) / (gend - gstart)

    # Check that pipeline output matches manual averaging results
    assert_allclose(manual_avg, avg_dark.data[:, 10, 10], rtol=1e-5)
    assert_allclose(manual_errs, avg_dark.err[:, 10, 10], rtol=1e-5)

    # Check that meta data was properly updated
    assert avg_dark.meta.exposure.nframes == nframes
    assert avg_dark.meta.exposure.ngroups == ngroups
    assert avg_dark.meta.exposure.groupgap == groupgap


def test_more_sci_frames(make_rampmodel, make_darkmodel):
    '''Check that data is unchanged if there are more frames in the science
    data is than in the dark reference file and verify that when the dark is
    not applied, the data is correctly flagged as such'''

    # size of integration
    ngroups = 7
    xsize = 200
    ysize = 200

    # create raw input data for step
    dm_ramp = make_rampmodel(ngroups, ysize, xsize)

    dm_ramp.meta.exposure.nframes = 1
    dm_ramp.meta.exposure.groupgap = 0

    # populate data array of science cube
    for i in range(0, ngroups-1):
        dm_ramp.data[0, i] = i

    refgroups = 5
    # create dark reference file model with fewer frames than science data
    dark = make_darkmodel(refgroups, ysize, xsize)

    # populate data array of reference file
    for i in range(0, refgroups - 1):
        dark.data[0, i] = i * 0.1

    # apply correction
    outfile = darkcorr(dm_ramp, dark)

    # test that the science data are not changed; input file = output file
    np.testing.assert_array_equal(dm_ramp.data, outfile.data)

    darkstatus = outfile.meta.cal_step.dark_sub
    assert darkstatus == 'SKIPPED'


def test_sub_by_frame(make_rampmodel, make_darkmodel):
    '''Check that if NFRAMES=1 and GROUPGAP=0 for the science data, the
    dark reference data are directly subtracted frame by frame'''

    # size of integration
    ngroups = 6
    xsize = 3
    ysize = 3

    # create raw input data for step
    dm_ramp = make_rampmodel(ngroups, ysize, xsize)

    dm_ramp.meta.exposure.nframes = 1
    dm_ramp.meta.exposure.groupgap = 0

    # populate data array of science cube
    for i in range(0, ngroups-1):
        dm_ramp.data[i] = i

    refgroups = 8
    dark = make_darkmodel(refgroups, ysize, xsize)

    # populate data array of reference file
    for i in range(0, refgroups - 1):
        dark.data[i] = i * 0.1

    # apply correction
    outfile = darkcorr(dm_ramp, dark)
    outdata = outfile.data

    diff = dm_ramp.data - dark.data[:ngroups]

    # test that the output data file is equal to the difference found when
    #     subtracting ref file from sci file
    np.testing.assert_array_equal(outdata, diff,
                                  err_msg='dark file should be subtracted'
                                          ' from sci file ')


def test_nan(make_rampmodel, make_darkmodel):
    '''Verify that when a dark has NaNs, these are correctly
    assumed as zero and the PIXELDQ is set properly'''

    # size of integration
    ngroups = 10
    xsize = 200
    ysize = 200

    # create raw input data for step
    dm_ramp = make_rampmodel(ngroups, ysize, xsize)

    dm_ramp.meta.exposure.nframes = 1
    dm_ramp.meta.exposure.groupgap = 0

    # populate data array of science cube
    for i in range(0, ngroups-1):
        dm_ramp.data[i, :, :] = i + 0.

    # create dark reference file model with more frames than science data
    refgroups = 15
    dark = make_darkmodel(refgroups, ysize, xsize)

    # populate data array of reference file
    for i in range(0, refgroups - 1):
        dark.data[0, i] = i * 0.1

    # set NaN in dark file
    dark.data[5, 100, 100] = np.nan

    # apply correction
    outfile = darkcorr(dm_ramp, dark)

    # test that the NaN dark reference pixel was set to 0 (nothing subtracted)

    assert outfile.data[5, 100, 100] == 5.0


def test_dq_combine(make_rampmodel, make_darkmodel):
    '''Verify that the DQ array of the dark is correctly combined
    with the PIXELDQ array of the science data.'''

    # size of integration
    ngroups = 5
    xsize = 200
    ysize = 200

    # create raw input data for step
    dm_ramp = make_rampmodel(ngroups, ysize, xsize)

    dm_ramp.meta.exposure.nframes = 1
    dm_ramp.meta.exposure.groupgap = 0

    # populate data array of science cube
    for i in range(1, ngroups-1):
        dm_ramp.data[i, :, :] = i

    # create dark reference file model with more frames than science data
    refgroups = 7
    dark = make_darkmodel(refgroups, ysize, xsize)

    jump_det = dqflags.pixel['JUMP_DET']
    saturated = dqflags.pixel['SATURATED']
    do_not_use = dqflags.pixel['DO_NOT_USE']

    # populate dq flags of sci pixeldq and reference dq
    dm_ramp.pixeldq[50, 50] = jump_det
    dm_ramp.pixeldq[50, 51] = saturated

    dark.dq[50, 50] = do_not_use
    dark.dq[50, 51] = do_not_use

    # run correction step
    outfile = darkcorr(dm_ramp, dark)

    # check that dq flags were correctly added
    assert outfile.pixeldq[50, 50] == np.bitwise_or(jump_det, do_not_use)
    assert outfile.pixeldq[50, 51] == np.bitwise_or(saturated, do_not_use)


def test_frame_avg(make_rampmodel, make_darkmodel):
    '''Check that if NFRAMES>1 or GROUPGAP>0, the frame-averaged dark data are
    subtracted group-by-group from science data groups and the ERR arrays are
    not modified'''

    # size of integration
    ngroups = 5
    xsize = 1032
    ysize = 1024

    # create raw input data for step
    dm_ramp = make_rampmodel(ngroups, ysize, xsize)

    dm_ramp.meta.exposure.nframes = 4
    dm_ramp.meta.exposure.groupgap = 0

    # populate data array of science cube
    for i in range(0, ngroups-1):
        dm_ramp.data[:, i] = i + 1

    # create dark reference file model
    refgroups = 20  # This needs to be 20 groups for the calculations to work
    dark = make_darkmodel(refgroups, ysize, xsize)

    # populate data array of reference file
    for i in range(0, refgroups - 1):
        dark.data[0, i] = i * 0.1

    # apply correction
    outfile = darkcorr(dm_ramp, dark)

    # dark frames should be averaged in groups of 4 frames
    # this will result in average values of 0.75, 0.775, 0.8, and 0.825
    # these values are then subtracted from frame values of 1, 2, 3 and 4
    assert outfile.data[0, 0, 0] == pytest.approx(0.25)
    assert outfile.data[0, 1, 0] == pytest.approx(1.225)
    assert outfile.data[0, 2, 0] == pytest.approx(2.2)
    assert outfile.data[0, 3, 0] == pytest.approx(3.175)

    # check that the error array is not modified.
    np.testing.assert_array_equal(outfile.err[:, :], 0)


@pytest.fixture(scope='function')
def make_rampmodel():
    '''Make Ramp model for testing'''

    def _ramp(ngroups, ysize, xsize):

        # create the data and groupdq arrays
        csize = (ngroups, ysize, xsize)

        data = np.full(csize, 1.0)

        # create a Roman datamodel for WFI data
        dm_ramp = RampModel(data=data)
        dm_ramp.meta.instrument.name = 'WFI'
        dm_ramp.meta.observation.datee = Time('2015-10-01T00:00:00')
        dm_ramp.meta.observation.time = Time('2015-10-01T01:01:00')
        dm_ramp.meta.description = 'Fake data.'

        return dm_ramp

    return _ramp


@pytest.fixture(scope='function')
def make_darkmodel():
    '''Make WFI dark model for testing'''

    def _dark(ngroups, ysize, xsize):
        # create the data and groupdq arrays
        csize = (ngroups, ysize, xsize)
        data = np.full(csize, 1.0)

        # create a dark datamodel for WFI data
        dark = DarkModel(data=data)
        dark.meta.instrument.name = 'WFI'

        dark.meta.exposure.nframes = 1
        dark.meta.exposure.groupgap = 0
        dark.meta.description = 'Fake data.'
        dark.meta.reftype = 'DARK'
        dark.meta.author = 'Alicia'
        dark.meta.pedigree = 'Dummy'
        dark.meta.useafter = Time('2015-10-01T00:00:00')

        return dark

    return _dark


@pytest.fixture(scope='function')
def setup_nrc_cube():
    '''Set up fake NIRCam data to test.'''

    def _cube(ma_tab_info, ngroups, nframes, groupgap, nrows, ncols):
        data_model = RampModel((ngroups, nrows, ncols))

        data_model.meta.exposure.ngroups = ngroups
        data_model.meta.exposure.groupgap = groupgap
        data_model.meta.exposure.nframes = nframes
        data_model.meta.exposure.frame_time = TFRAME
        data_model.meta.exposure.group_time = (nframes + groupgap) * TFRAME
        data_model.meta.instrument.name = 'WFI'
        data_model.meta.instrument.detector = 'WFI01'

        data_model.meta.observation.date = Time('2017-11-01T00:00:00')
        data_model.meta.observation.time = Time('2017-11-01T05:13:00')

        dark_model = DarkModel((NGROUPS_DARK, 2048, 2048))

        dark_model.meta.exposure.ngroups = NGROUPS_DARK
        dark_model.meta.exposure.groupgap = 0
        dark_model.meta.exposure.nframes = 1
        dark_model.meta.instrument.name = 'WFI'
        dark_model.meta.description = 'Fake data.'
        dark_model.meta.telescope = 'ROMAN'
        dark_model.meta.reftype = 'DARK'
        dark_model.meta.author = 'Alicia'
        dark_model.meta.pedigree = 'Dummy'
        dark_model.meta.useafter = Time('2016-10-01T00:00:00')

        return data_model, dark_model

    return _cube



"""Base class for Roman data models"""

from stdatamodels import DataModel
from astropy.time import Time


class RomanDataModel(DataModel):
    """
    Base class for Roman data models.  Also serves as a generic
    model class for data that does not identify its model type.
    """
    schema_url = "http://stsci.edu/schemas/roman_datamodel/core.schema"

    # CRDS observatory code for Roman.  Used by stpipe when selecting
    # references.
    crds_observatory = "roman"

    def get_crds_parameters(self):
        """
        Get parameters used by CRDS to select references for this model.

        Returns
        -------
        dict
        """
        return {
            key: val for key, val in self.to_flat_dict(include_arrays=False).items()
            if isinstance(val, (str, int, float, complex, bool))
        }

    def on_init(self, init):
        """
        Hook invoked by the base class before returning a newly
        created model instance.
        """
        super().on_init(init)

        if self.meta.date is None:
            self.meta.date = Time(Time.now(), format="isot")

    def on_save(self, init):
        """
        Hook invoked by the base class before writing a model
        to a file.
        """
        super().on_save(init)

        self.meta.date = Time(Time.now(), format="isot")



"""Module for dark subtracting science data sets"""
#
#  Module for dark subtracting science data sets
#

import logging
import warnings
import numpy as np

from .. import datamodels

log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)


def do_correction(input_model, dark_model, dark_output=None):
    """
    Short Summary
    -------------
    Execute all tasks for Dark Current Subtraction

    Parameters
    ----------
    input_model: data model object
        science data to be corrected

    dark_model: dark model object
        dark data

    dark_output: string
        file name in which to optionally save averaged dark data

    Returns
    -------
    output_model: data model object
        dark-subtracted science data

    """
    # Save some data params for easy use later
    sci_ngroups = input_model.data.shape[0]
    sci_nframes = input_model.meta.exposure.nframes
    sci_groupgap = input_model.meta.exposure.groupgap

    drk_ngroups = dark_model.data.shape[0]
    drk_nframes = dark_model.meta.exposure.nframes
    drk_groupgap = dark_model.meta.exposure.groupgap

    log.info(
        'Science data ngroups=%d, nframes=%d, groupgap=%d',
        sci_ngroups, sci_nframes, sci_groupgap
    )
    log.info(
        'Dark data ngroups=%d, nframes=%d, groupgap=%d',
        drk_ngroups, drk_nframes, drk_groupgap
    )

    # Check that the number of groups in the science data does not exceed
    # the number of groups in the dark current array.
    sci_total_frames = sci_ngroups * (sci_nframes + sci_groupgap)
    drk_total_frames = drk_ngroups * (drk_nframes + drk_groupgap)
    if sci_total_frames > drk_total_frames:
        log.warning(
            "Not enough data in dark reference file to match to "
            "science data."
        )
        log.warning("Input will be returned without subtracting dark current.")
        input_model.meta.cal_step.dark_sub = 'SKIPPED'
        return input_model.copy()

    # Check that the value of nframes and groupgap in the dark
    # are not greater than those of the science data
    if drk_nframes > sci_nframes or drk_groupgap > sci_groupgap:
        log.warning(
            "The value of nframes or groupgap in the dark data is "
            "greater than that of the science data."
            "Input will be returned without subtracting dark current."
        )
        input_model.meta.cal_step.dark_sub = 'SKIPPED'
        return input_model.copy()

    # Replace NaN's in the dark with zeros
    dark_model.data[np.isnan(dark_model.data)] = 0.0

    # Check whether the dark and science data have matching
    # nframes and groupgap settings.
    if sci_nframes == drk_nframes and sci_groupgap == drk_groupgap:

        # They match, so we can subtract the dark ref file data directly
        output_model = subtract_dark(input_model, dark_model)

        # If the user requested to have the dark file saved,
        # save the reference model as this file. This will
        # ensure consistency from the user's standpoint
        if dark_output is not None:
            log.info('Writing dark current data to %s', dark_output)
            dark_model.save(dark_output)

    else:

        # Create a frame-averaged version of the dark data to match
        # the nframes and groupgap settings of the science data.
        averaged_dark = average_dark_frames(
            dark_model, sci_ngroups, sci_nframes, sci_groupgap)

        # Save the frame-averaged dark data that was just created,
        # if requested by the user
        if dark_output is not None:
            log.info('Writing dark current data to %s', dark_output)
            averaged_dark.save(dark_output)

        # Subtract the frame-averaged dark data from the science data
        output_model = subtract_dark(input_model, averaged_dark)

        averaged_dark.close()

    output_model.meta.cal_step.dark_sub = 'COMPLETE'

    return output_model


def average_dark_frames(input_dark, ngroups, nframes, groupgap):
    """
    Averages the individual frames of data in a dark reference
    file to match the group structure of a science data set.

    Parameters
    ----------
    input_dark: dark data model
        the input dark data

    ngroups: int
        number of groups in the science data set

    nframes: int
        number of frames per group in the science data set

    groupgap: int
        number of frames skipped between groups in the science data set

    Returns
    -------
    avg_dark: dark data model
        New dark object with averaged frames

    """
    # Create a model for the averaged dark data
    dny = input_dark.data.shape[1]
    dnx = input_dark.data.shape[2]
    avg_dark = datamodels.DarkModel((ngroups, dny, dnx))

    # 03/12/21 this block is a workaround for datamodel's update bug
    input_dark.meta.instrument.detector = 'WFI01'
    avg_dark.meta.instrument.name = 'WFI'
    avg_dark.meta.instrument.detector = 'WFI01'
    avg_dark.meta.update(input_dark.meta._instance)
    avg_dark.update(input_dark)

    # Do a direct copy of the 2-d DQ array into the new dark
    avg_dark.dq = input_dark.dq

    # Loop over the groups of the input science data, copying or
    # averaging the dark frames to match the group structure
    start = 0

    for group in range(ngroups):
        end = start + nframes

        # If there's only 1 frame per group, just copy the dark frames
        if nframes == 1:
            log.debug('copy dark frame %d', start)
            avg_dark.data[group] = input_dark.data[start]
            avg_dark.err[group] = input_dark.err[start]

        # Otherwise average nframes into a new group: take the mean of
        # the SCI arrays and the quadratic sum of the ERR arrays.
        else:
            log.debug('average dark frames %d to %d', start + 1, end)

            with warnings.catch_warnings():
                warnings.simplefilter("ignore", category=RuntimeWarning)
                avg_dark.data[group] = input_dark.data[start:end].mean(axis=0)

            avg_dark.err[group] = np.sqrt(np.add.reduce(
                input_dark.err[start:end]**2, axis=0)) / (end - start)

        # Skip over unused frames
        start = end + groupgap

    # Reset some metadata values for the averaged dark
    avg_dark.meta.exposure.nframes = nframes
    avg_dark.meta.exposure.ngroups = ngroups
    avg_dark.meta.exposure.groupgap = groupgap

    return avg_dark


def subtract_dark(input, dark):
    """
    Subtracts dark current data from science arrays, combines and updates data
    quality array based on DQ flags in the dark arrays.

    Parameters
    ----------
    input: data model object
        the input science data

    dark: dark model object
        the dark current data

    Returns
    -------
    output: data model object
        dark-subtracted science data

    """
    log.debug("subtract_dark: ngroups=%d, size=%d,%d",
              input.data.shape[0], input.data.shape[1], input.data.shape[2])

    # Create output as a copy of the input science data model
    output = input.copy()

    # Combine the dark and science DQ data
    output.pixeldq = np.bitwise_or(input.pixeldq, dark.dq)

    # loop over all groups in input science data
    for j in range(input.data.shape[0]):
        # subtract the SCI arrays
        output.data[j] -= dark.data[j]

    return output



import pytest

@pytest.fixture(autouse=True)
def _docdir(request):

    # Trigger ONLY for doctestplus.
    doctest_plugin = request.config.pluginmanager.getplugin("doctestplus")
    if isinstance(request.node.parent, doctest_plugin._doctest_textfile_item_cls):
        tmpdir = request.getfixturevalue('tmpdir')
        with tmpdir.as_cwd():
            yield
    else:
        yield



from ..stpipe import RomanStep
from .. import datamodels
from . import dark_sub


__all__ = ["DarkCurrentStep"]


class DarkCurrentStep(RomanStep):
    """
    DarkCurrentStep: Performs dark current correction by subtracting
    dark current reference data from the input science data model.
    """

    spec = """
        dark_output = output_file(default = None) # Dark corrected model
    """

    reference_file_types = ['dark']

    def process(self, input):

        # Open the input data model
        with datamodels.RampModel(input) as input_model:

            # Get the name of the dark reference file to use
            self.dark_name = self.get_reference_file(input_model, 'dark')
            self.log.info('Using DARK reference file %s', self.dark_name)

            # Check for a valid reference file
            if self.dark_name == 'N/A':
                self.log.warning('No DARK reference file found')
                self.log.warning('Dark current step will be skipped')
                result = input_model.copy()
                result.meta.cal_step.dark = 'SKIPPED'
                return result

            # Create name for the intermediate dark, if desired.
            dark_output = self.dark_output
            if dark_output is not None:
                dark_output = self.make_output_path(
                    None, basepath=dark_output, ignore_use_model=True
                )

            dark_model = datamodels.DarkModel(self.dark_name)

            # Do the dark correction
            result = dark_sub.do_correction(
                input_model, dark_model, dark_output
            )
            dark_model.close()

        return result



import os
import pytest

from romancal.stpipe import RomanStep
from romancal.step import FlatFieldStep
import roman_datamodels as rdm
from .regtestdata import compare_asdf


@pytest.mark.bigdata
def test_flat_field_image_step(rtdata, ignore_asdf_paths):

    rtdata.get_data("WFI/image/l2_0001_rate.asdf")
    rtdata.input = "l2_0001_rate.asdf"

    # Test CRDS
    step = FlatFieldStep()
    model = rdm.open(rtdata.input)
    ref_file_path = step.get_reference_file(model, "flat")
    ref_file_name = os.path.split(ref_file_path)[-1]
    assert "roman_wfi_flat" in ref_file_name

    # Test FlatFieldStep
    output = "l2_0001_rate_flatfieldstep.asdf"
    rtdata.output = output
    args = ["romancal.step.FlatFieldStep", rtdata.input]
    RomanStep.from_cmdline(args)
    rtdata.get_truth(f"truth/WFI/image/{output}")
    compare_asdf(rtdata.output, rtdata.truth, **ignore_asdf_paths)

@pytest.mark.bigdata
def test_flat_field_grism_step(rtdata, ignore_asdf_paths):

    rtdata.get_data("WFI/grism/l2_0001_grism_rate.asdf")
    rtdata.input = "l2_0001_grism_rate.asdf"

    # Test CRDS
    step = FlatFieldStep()
    model = rdm.open(rtdata.input)
    ref_file_path = step.get_reference_file(model, "flat")
    ref_file_name = os.path.split(ref_file_path)[-1]
    assert "roman_wfi_flat" in ref_file_name

    # Test FlatFieldStep
    output = "l2_0001_grism_rate_flatfieldstep.asdf"
    rtdata.output = output
    args = ["romancal.step.FlatFieldStep", rtdata.input]
    RomanStep.from_cmdline(args)
    rtdata.get_truth(f"truth/WFI/grism/{output}")
    compare_asdf(rtdata.output, rtdata.truth, **ignore_asdf_paths)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import sys
from pkg_resources import get_distribution, DistributionNotFound
if sys.version_info < (3, 6):
    raise ImportError("Romancal supports Python versions 3.6 and above.")  # pragma: no cover

try:
    __version__ = get_distribution(__name__).version
except DistributionNotFound:  # pragma: no cover
    # package is not installed
    pass  # pragma: no cover



import os
import glob
import io


S3_BUCKET_NAME = "test-s3-data"


class MockS3Client:
    def __init__(self, s3_test_data_path):
        self.s3_test_data_path = s3_test_data_path

    def get_object(self, bucket_name, key):
        assert self.object_exists(bucket_name, key)

        with open(self._get_path(key), "rb") as f:
            return io.BytesIO(f.read())

    def object_exists(self, bucket_name, key):
        if bucket_name != S3_BUCKET_NAME:
            return False

        return os.path.isfile(self._get_path(key))

    def prefix_exists(self, bucket_name, key_prefix):
        return any(self.iterate_keys(bucket_name, key_prefix))

    def iterate_keys(self, bucket_name, key_prefix):
        if bucket_name != S3_BUCKET_NAME:
            return

        for k in self._list_keys():
            if k.startswith(key_prefix):
                yield k

    def _get_path(self, key):
        return os.path.join(self.s3_test_data_path, key)

    def _list_keys(self):
        paths = glob.glob(self.s3_test_data_path + "/**", recursive=True)
        paths = [p for p in paths if os.path.isfile(p)]
        return [p.replace(self.s3_test_data_path, "") for p in paths]



"""
A data model for 3D ramps
"""
from .core import RomanDataModel

__all__ = ["RampModel"]


class RampModel(RomanDataModel):
    """
    A data model for 3D ramps.

    Parameters
    __________
    data : numpy float32 array
         The science data

    pixeldq : numpy uint32 array
         2-D data quality array for all planes

    groupdq : numpy uint8 array
         3-D data quality array for each plane

    err : numpy float32 array
         Error array

    group : numpy table
         group parameters table
    """
    schema_url = "http://stsci.edu/schemas/roman_datamodel/ramp.schema"

    def __init__(self, init=None, **kwargs):
        super(RampModel, self).__init__(init=init, **kwargs)

        # Implicitly create arrays
        self.pixeldq = self.pixeldq
        self.groupdq = self.groupdq
        self.err = self.err



"""
Implementation of the datamodels.open function.
"""
from pathlib import PurePath, Path

import asdf

from .core import RomanDataModel
from .reference_files.referencefile import ReferenceFileModel
from .reference_files.flat import FlatModel


def open(init, memmap=False, **model_kwargs):
    """
    Open an ASDF file and wrap it in a data model class.

    Parameters
    ----------
    init : str or pathlib.PurePath or asdf.AsdfFile or
           romancal.datamodels.RomanDataModel
        If str or `~pathlib.PurePath`, filesystem path to an ASDF file.
        If `~asdf.AsdfFile`, an open ASDF file.
        If `~romancal.datamodels.RomanDataModel`, an already open model.

    memmap : bool, optional
        Set to True to enable memory-mapping of arrays.  Ignored if the
        ``init`` argument is an AsdfFile.

    model_kwargs : dict, optional
        Additional arguments to pass when initializing the model.

    Returns
    -------
    model : RomanDataModel
        Instance of `RomanDataModel` or subclass.
    """
    if isinstance(init, (str, PurePath)):
        if isinstance(init, str):
            init = Path(init)

        if not init.is_file():
            raise FileNotFoundError(f"File at path '{init}' does not exist")

        asdf_kwargs = {
            # The copy_arrays argument instructs the asdf library to
            # make in-memory copies of arrays, so we need to send it
            # the opposite of our memmap value.
            "copy_arrays": not memmap
        }

        asdf_file = asdf.open(init, **asdf_kwargs)
    elif isinstance(init, asdf.AsdfFile):
        asdf_file = init
    elif isinstance(init, RomanDataModel):
        # Make a copy of the model so that the original instance can
        # be closed without impacting this one.
        return init.__class__(init)
    else:
        raise TypeError("init must be a path to an ASDF file, an open ",
                        "AsdfFile instance or a RomanDataModel")

    model_class = _select_model_class(asdf_file)

    model = model_class(asdf_file, **model_kwargs)

    return model


def _select_model_class(asdf_file):
    """
    Select a model class based on ASDF file metadata.

    Parameters
    ----------
    asdf_file : asdf.AsdfFile

    Returns
    -------
    type
        `RomanDataModel` or subclass.
    """
    # TODO: Still need to decide how this is going to work.
    # Options include:
    # - Choose according to model_type metadata field (similar to jwst)
    # - Infer from other relevant metadata (reference file type, exposure
    #   type, etc)
    # - Use ASDF tags to automatically instantiate the correct model
    # - Use RomanDataModel for all data but select the schema according to
    #    one of the above
    # - ???

    # Check for the existence of reftype to indicate a reference file model
    reftype = asdf_file["meta"].get("reftype")
    if reftype is not None:
        # Check if flat file model
        if reftype == "FLAT":
            return FlatModel
        # Return base reference file model
        return ReferenceFileModel

    # Not a reference file model
    return RomanDataModel



from .referencefile import ReferenceFileModel
from ..dynamicdq import dynamic_mask


class DarkModel(ReferenceFileModel):
    """
    A data model for dark reference files.

    Parameters
    __________
    data : numpy float32 array
         Dark current array

    dq : numpy uint16 array
         2-D data quality array for all planes

    err : numpy float32 array
         Error array
    """
    schema_url = "http://stsci.edu/schemas/roman_datamodel/reference_files/dark.schema"

    def __init__(self, init=None, **kwargs):
        super().__init__(init=init, **kwargs)

        # Saved for DQ code implementation
        self.dq = dynamic_mask(self)

        # Implicitly create arrays
        self.dq = self.dq
        self.err = self.err



# #############################################################
# Setup a base class and instantiate it in order to provide the
# SDP pool lists for test parametrization.
# #############################################################
from romancal.tests.base_classes import BaseRomanTest


class SDPPoolsSource(BaseRomanTest):
    """Retreive the SDP pools from the test data source

    These tests are very much tied to using `pytest` and the `ci-watson` plugin.
    In particular, there are references to `pytest.config` that only exist when
    running under pytest. Such references are stubbed out with best defaults used.
    """

    # The following will be set during both plugin instantiation
    # of `BaseJWSTTest`, or explicitly during dynamic parametrization
    # The empty string defaults indicate that data source roots have
    # been defined.
    inputs_root = ''
    results_root = ''

    input_loc = 'associations'
    test_dir = 'sdp'
    ref_loc = [test_dir, 'truth']

    _pool_paths = None
    _truth_paths = None

    @property
    def pool_paths(self):
        """Get the association pools"""
        if self._pool_paths is None:
            self._pool_paths = self.data_glob(self.test_dir, 'pools', glob='*.csv')
        return self._pool_paths

    @property
    def truth_paths(self):
        """Get the truth associations"""
        if self._truth_paths is None:
            self._truth_paths = self.data_glob(*self.ref_loc, glob='*.json')
        return self._truth_paths



"""
A data model for post ramp fitted data
"""
from .core import RomanDataModel

__all__ = ["RampFitOutputModel"]


class RampFitOutputModel(RomanDataModel):
    """
    A data model for the optional output of the ramp fitting step.

    In the parameter definitions below, `max_seg` is the maximum number
    of segments that were fit, and `ny` and `nx` are the height and width
    of the image.

    Parameters
    __________

    slope : numpy float32 array (max_seg, ny, nx)
        Segment-specific slope

    sigslope : numpy float32 array (max_seg, ny, nx)
        Sigma for segment-specific slope

    var_poisson : numpy float32 array (max_seg, ny, nx)
        Variance due to poisson noise for segment-specific slope

    var_rnoise : numpy float32 array (max_seg, ny, nx)
        Variance due to read noise for segment-specific slope

    yint : numpy float32 array (max_seg, ny, nx)
        Segment-specific y-intercept

    sigyint : numpy float32 array (max_seg, ny, nx)
        Sigma for segment-specific y-intercept

    pedestal : numpy float32 array (max_seg, ny, nx)
        Pedestal array

    weights : numpy float32 array (max_seg, ny, nx)
        Weights for segment-specific fits

    crmag : numpy float32 array (max_seg, ny, nx)
        Approximate CR magnitudes
    """
    schema_url = ("http://stsci.edu/schemas/roman_datamodel/"
                  "rampfitoutput.schema")



# Base Roman data model
from .core import RomanDataModel

# Reference File data models
from .reference_files.referencefile import ReferenceFileModel
from .reference_files.flat import FlatModel
from .reference_files.dark import DarkModel
from .reference_files.gain import GainModel
from .reference_files.mask import MaskModel
from .reference_files.readnoise import ReadNoiseModel

# Image Models
from .gls_rampfit import GLS_RampFitModel
from .image import ImageModel
from .level1 import Level1FileModel
from .ramp import RampModel
from .rampfitoutput import RampFitOutputModel

from .open_impl import open


__all__ = ["DarkModel",
           "FlatModel",
           "GainModel",
           "GLS_RampFitModel",
           "ImageModel",
           "Level1FileModel",
           "MaskModel",
           "RampModel",
           "ReadNoiseModel",
           "RampFitOutputModel",
           "ReferenceFileModel",
           "RomanDataModel",
           "open"]



""" Set DQ mask dynamically

    Parameters
    -----------
    input : model
        Filename for a FITS image or association table, or a `CubeModel`.

    output: mask
        data quality mask

"""

import logging
import numpy as np
from . import dqflags

log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)
log.addHandler(logging.NullHandler())


def dynamic_mask(input_model):
    """Return a mask model given a mask with dynamic DQ flags """
    # Return a mask model given a mask with dynamic DQ flags
    # Dynamic flags define what each plane refers to using the DQ_DEF extension

    dq_table = input_model.dq_def
    # Get the DQ array and the flag definitions
    if (dq_table is not None and not np.isscalar(dq_table) and
            not dq_table.shape and not dq_table):
        #
        # Make an empty mask
        dqmask = np.zeros(input_model.dq.shape, dtype=input_model.dq.dtype)
        for record in dq_table:
            bitplane = record['VALUE']
            dqname = record['NAME'].strip()
            try:
                standard_bitvalue = dqflags.pixel[dqname]
            except KeyError:
                log.warning('Keyword %s does not correspond to an existing'
                            'DQ mnemonic so will be ignored {}', dqname)
                continue
            just_this_bit = np.bitwise_and(input_model.dq, bitplane)
            pixels = np.where(just_this_bit != 0)
            dqmask[pixels] = np.bitwise_or(dqmask[pixels], standard_bitvalue)
    else:
        dqmask = input_model.dq

    return dqmask



from .referencefile import ReferenceFileModel


class GainModel(ReferenceFileModel):
    """
    A data model for 2D gain.

    Parameters
    __________
    data : numpy float32 array
         The gain
    """
    schema_url = "http://stsci.edu/schemas/roman_datamodel/reference_files/gain.schema"



"""
Data model for level 1 data files
"""

from .core import RomanDataModel


class Level1FileModel(RomanDataModel):
    """
    Data model for level 1 data files.

    Parameters
    __________
    data : numpy float32 array
         The science data

    dq : numpy uint32 array
         Data quality array

    err : numpy float32 array
         Error array

    var_poisson : numpy float32 array
         variance due to poisson noise

    var_rnoise : numpy float32 array
         variance due to read noise

    area : numpy float32 array
         Pixel area map array

    """
    schema_url = "http://stsci.edu/schemas/roman_datamodel/level1.schema"

    def __init__(self, init=None, **kwargs):
        super().__init__(init=init, **kwargs)



from .referencefile import ReferenceFileModel
from ..dynamicdq import dynamic_mask


class MaskModel(ReferenceFileModel):
    """
    A data model for 2D masks.

    Parameters
    __________
    dq : numpy uint32 array
         The mask

    dq_def : numpy table
         DQ flag definitions
    """
    schema_url = "http://stsci.edu/schemas/roman_datamodel/reference_files/mask.schema"

    def __init__(self, init=None, **kwargs):
        super(MaskModel, self).__init__(init=init, **kwargs)

        if self.dq is not None or self.dq_def is not None:
            self.dq = dynamic_mask(self)

        # Implicitly create arrays
        self.dq = self.dq

    def get_primary_array_name(self):
        """
        Returns the name "primary" array for this model, which
        controls the size of other arrays that are implicitly created.
        This is intended to be overridden in the subclasses if the
        primary array's name is not "data".
        """
        return 'dq'



from .referencefile import ReferenceFileModel


class ReadNoiseModel(ReferenceFileModel):
    """
    A data model for 2D read noise.

    Parameters
    __________
    data : numpy float32 array
         The read noise
    """
    schema_url = "http://stsci.edu/schemas/roman_datamodel/reference_files/readnoise.schema"



"""Extension that maps datamodel schema URIs to their corresponding disk
locations """

from pathlib import Path

from asdf.extension import AsdfExtension
from asdf import util


SCHEMAS_ROOT = (Path(__file__).parent/"schemas").resolve()
URI_PREFIX = "http://stsci.edu/schemas/roman_datamodel/"


class RomanDataModelExtension(AsdfExtension):
    """
    Extension that maps datamodel schema URIs to their corresponding
    locations on the disk.  This allows the asdf package to locate
    the schema content when we request validation against one of these
    URIs.
    """

    # The AsdfExtension interface requires types and tag_mapping attributes,
    # but we won't actually be using them here.
    types = []
    tag_mapping = []

    url_mapping = [(URI_PREFIX,
                    util.filepath_to_url(str(SCHEMAS_ROOT)) +
                    "/{url_suffix}.yaml"), ]



#! /usr/bin/env python
#
import logging
import numpy as np

from ..stpipe import Step
from .. import datamodels
from . import ramp_fit

from . import JWST   # 020821 syntax ?
from . import Roman   # 020821 syntax ?

log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)


__all__ = ["RampFitStep"]


class RampFitStep (Step):

    """
    This step fits a straight line to the value of counts vs. time to
    determine the mean count rate for each pixel.
    """

    spec = """
        opt_name = string(default='')
        maximum_cores = option('none','quarter','half','all',default='none')
            # max number of processes to create
    """

    # Prior to 04/26/17, the following were also in the spec above:
    #      algorithm = option('OLS', 'GLS', default='OLS') # 'OLS' or 'GLS'
    #      weighting = option('unweighted', 'optimal', default='unweighted') \
    #      # 'unweighted' or 'optimal'
    # As of 04/26/17, the only allowed algorithm is 'ols', and the
    #      only allowed weighting is 'optimal'.

    algorithm = 'ols'      # Only algorithm allowed ?
#    algorithm = 'gls'

    weighting = 'optimal'  # Only weighting allowed ?

    reference_file_types = ['readnoise', 'gain']

    def process(self, input):

        # Open input as a Roman DataModel (single integration; 3D arrays)
        with Roman.datamodels.RampModel(input) as R_input_model:  # syntax ?

            # Create a JWST DataModel (4D arrays) to populate
            input_model = JWST.datamodels.RampModel()  # syntax ?

            # Populate JWST model with correct sized arrays from Roman model
            input_model.meta = R_input_model.meta
            input_model.data = np.broadcast_to(R_input_model.data, (1,) +
                                               R_input_model.data.shape)

            input_model.pixeldq = R_input_model.pixeldq

            input_model.groupdq = np.broadcast_to(R_input_model.groupdq, (1,) +
                                                  R_input_model.groupdq.shape)

            input_model.err = np.broadcast_to(R_input_model.err, (1,) +
                                              R_input_model.err.shape)

            input_model.refout = np.broadcast_to(R_input_model.refout, (1,) +
                                                 R_input_model.refout.shape)

            max_cores = self.maximum_cores
            readnoise_filename = self.get_reference_file(input_model,
                                                         'readnoise')
            gain_filename = self.get_reference_file(input_model, 'gain')

            log.info('Using READNOISE reference file: %s', readnoise_filename)
            readnoise_model = datamodels.ReadnoiseModel(readnoise_filename)
            log.info('Using GAIN reference file: %s', gain_filename)
            gain_model = datamodels.GainModel(gain_filename)

            # Try to retrieve the gain factor from the gain reference file.
            # If found, store it in the science model meta data, so that it's
            # available later in the gain_scale step, which avoids having to
            # load the gain ref file again in that step.
            if gain_model.meta.exposure.gain_factor is not None:
                input_model.meta.exposure.gain_factor = \
                    gain_model.meta.exposure.gain_factor

            log.info('Using algorithm = %s' % self.algorithm)
            log.info('Using weighting = %s' % self.weighting)

            buffsize = ramp_fit.BUFSIZE
            if self.algorithm == "GLS":
                buffsize //= 10

            out_model, int_model, opt_model, gls_opt_model = ramp_fit.ramp_fit(
                input_model, buffsize,
                readnoise_model, gain_model, self.algorithm,
                self.weighting, max_cores
            )
            # The above int_model is None

            # Remove the empty 0th dimension from the JWST model and insert
            # the results into the Roman DataModel
            R_input_model.data = input_model.data[-1]
            R_input_model.groupdq = input_model.groupdq[-1]
            R_input_model.err = input_model.err[-1]
            R_input_model.refout = input_model.refout[-1]

            readnoise_model.close()
            gain_model.close()

        # Save the OLS optional fit product, if it exists
        if opt_model is not None:
            self.save_model(opt_model, 'fitopt', output_file=self.opt_name)

        # Save the GLS optional fit product, if it exists
        if gls_opt_model is not None:
            self.save_model(
                gls_opt_model, 'fitoptgls', output_file=self.opt_name
            )

        if out_model is not None:
            out_model.meta.cal_step.ramp_fit = 'COMPLETE'

        return out_model, None  # 'None' for int_model



"""
Roman Calibration Pipeline base class
"""
from stpipe import Step, Pipeline

import roman_datamodels as rdm


class RomanStep(Step):
    """
    Base class for Roman calibration pipeline steps.
    """
    spec = """
    output_ext =  string(default='.asdf')    # Default type of output
    """

    @classmethod
    def _datamodels_open(cls, init, **kwargs):
        """
        Provide access to this package's datamodels.open function
        so that the stpipe infrastructure knows how to instantiate
        models.
        """
        return rdm.open(init, **kwargs)

    def finalize_result(self, model, reference_files_used):
        """
        Hook that allows the Step to set metadata on the output model
        before save.

        Parameters
        ----------
        model : stdatamodels.DataModel
            Output model.

        reference_files_used : list of tuple(str, str)
            List of reference files used.  The first element of each tuple
            is the reftype code, the second element is the filename.
        """
        #  JWST uses this to add the cal code and CRDS software versions.
        pass

    def record_step_status(self, model, step_name, success=True):
        """
        Record step completion status in the model's metadata.

        Parameters
        ----------
        model : stdatamodels.DataModel
            Output model.
        step_name : str
            Calibration step name.
        success : bool
            If True, then the step was run successfully.
        """
        # JWST sets model.meta.cal_step.<step name> here.  Roman
        # may do the same, depending on how the metadata format
        # turns out.  Seems like we might be able to combine this
        # with finalize_result somehow.
        pass

    def remove_suffix(self, name):
        """
        Remove any Roman step-specific suffix from the given filename.

        Parameters
        ----------
        name : str
            Filename.

        Returns
        -------
        str
            Filename with step suffix removed.
        """
        # JWST maintains a list of relevant suffixes that is monitored
        # by tests to be up-to-date.  Roman will likely need to do
        # something similar.
        return name, "_"


# RomanPipeline needs to inherit from Pipeline, but also
# be a subclass of RomanStep so that it will pass checks
# when constructing a pipeline using RomanStep class methods.
class RomanPipeline(Pipeline, RomanStep):
    pass



""" Roman Data Quality Flags

The definitions are documented in the Roman RTD:

[NOTE: Documentation not yet implemented. Fix this URL when completed.]

https://roman-cal-pipeline.readthedocs.io/en/latest/roman/references_general/references_general.html#data-quality-flags


Implementation
-------------

The flags are implemented as "bit flags": Each flag is assigned a bit position
in a byte, or multi-byte word, of memory. If that bit is set, the flag assigned
to that bit is interpreted as being set or active.

The data structure that stores bit flags is just the standard Python `int`,
which provides 32 bits. Bits of an integer are most easily referred to using
the formula `2**bit_number` where `bit_number` is the 0-index bit of interest.
"""
from astropy.nddata.bitmask import interpret_bit_flags as ap_interpret_bit_flags

from romancal.lib.basic_utils import multiple_replace

# Pixel-specific flags
pixel = {'GOOD':             0,      # No bits set, all is good
         'DO_NOT_USE':       2**0,   # Bad pixel. Do not use.
         'SATURATED':        2**1,   # Pixel saturated during exposure
         'JUMP_DET':         2**2,   # Jump detected during exposure
         'DROPOUT':          2**3,   # Data lost in transmission
         'RESERVED_1':       2**4,   #
         'PERSISTENCE':      2**5,   # High persistence (was RESERVED_2)
         'RESERVED_3':       2**6,   #
         'RESERVED_4':       2**7,   #
         'UNRELIABLE_ERROR': 2**8,   # Uncertainty exceeds quoted error
         'NON_SCIENCE':      2**9,   # Pixel not on science portion of detector
         'DEAD':             2**10,  # Dead pixel
         'HOT':              2**11,  # Hot pixel
         'WARM':             2**12,  # Warm pixel
         'LOW_QE':           2**13,  # Low quantum efficiency
         'TELEGRAPH':        2**15,  # Telegraph pixel
         'NONLINEAR':        2**16,  # Pixel highly nonlinear
         'BAD_REF_PIXEL':    2**17,  # Reference pixel cannot be used
         'NO_FLAT_FIELD':    2**18,  # Flat field cannot be measured
         'NO_GAIN_VALUE':    2**19,  # Gain cannot be measured
         'NO_LIN_CORR':      2**20,  # Linearity correction not available
         'NO_SAT_CHECK':     2**21,  # Saturation check not available
         'UNRELIABLE_BIAS':  2**22,  # Bias variance large
         'UNRELIABLE_DARK':  2**23,  # Dark variance large
         'UNRELIABLE_SLOPE': 2**24,  # Slope variance large (i.e., noisy pixel)
         'UNRELIABLE_FLAT':  2**25,  # Flat variance large
         'RESERVED_5':       2**26,  #
         'RESERVED_6':       2**27,  #
         'UNRELIABLE_RESET': 2**28,  # Sensitive to reset anomaly
         'RESERVED_7':       2**29,  #
         'OTHER_BAD_PIXEL':  2**30,  # A catch-all flag
         'REFERENCE_PIXEL':  2**31,  # Pixel is a reference pixel
         }


# Group-specific flags. Once groups are combined, these flags
# are equivalent to the pixel-specific flags.
group = {'GOOD':       pixel['GOOD'],
         'DO_NOT_USE': pixel['DO_NOT_USE'],
         'SATURATED':  pixel['SATURATED'],
         'JUMP_DET':   pixel['JUMP_DET'],
         'DROPOUT':    pixel['DROPOUT'],
         }


def interpret_bit_flags(bit_flags, flip_bits=None):
    """Converts input bit flags to a single integer value (bit mask) or `None`.

    Wraps `astropy.nddate.bitmask.interpret_bit_flags`, allowing the Roman
    bit mnemonics to be used in place of integers.

    Parameters
    ----------
    bit_flags : int, str, list, None
        See `astropy.nddate.bitmask.interpret_bit_flags`.
        Also allows strings using Roman mnemonics

    flip_bits : bool, None
        See `astropy.nddata.bitmask.interpret_bit_flags`.

    Returns
    -------
    bitmask : int or None
        Returns an integer bit mask formed from the input bit value or `None`
        if input ``bit_flags`` parameter is `None` or an empty string.
        If input string value was prepended with '~' (or ``flip_bits`` was set
        to `True`), then returned value will have its bits flipped
        (inverse mask).

    Examples
    --------
    Using Roman mnemonics:
    TBD
    """
    bit_flags_dm = bit_flags
    if isinstance(bit_flags, str):
        dm_flags = {
            key: str(val)
            for key, val in pixel.items()
        }
        bit_flags_dm = multiple_replace(bit_flags, dm_flags)

    return ap_interpret_bit_flags(bit_flags_dm, flip_bits=flip_bits)


def dqflags_to_mnemonics(dqflags, mnemonic_map=pixel):
    """Interpret value as bit flags and return the mnemonics

    Parameters
    ----------
    dqflags : int-like
        The value to interpret as DQ flags

    mnemonic_map: {str: int[,...]}
        Dictionary associating the mnemonic string to an integer value
        representing the set bit for that mnemonic.

    Returns
    -------
    mnemonics : {str[,...]}
        Set of mnemonics represented by the set bit flags

    Examples
    --------
    >>> dqflags_to_mnemonics(1)
    {'DO_NOT_USE'}

    >>> dqflags_to_mnemonics(7)             #doctest: +SKIP
    {'JUMP_DET', 'DO_NOT_USE', 'SATURATED'}

    >>> dqflags_to_mnemonics(7) == {'JUMP_DET', 'DO_NOT_USE', 'SATURATED'}
    True

    >>> dqflags_to_mnemonics(1, mnemonic_map=pixel)
    {'DO_NOT_USE'}

    >>> dqflags_to_mnemonics(1, mnemonic_map=group)
    {'DO_NOT_USE'}
    """
    mnemonics = {
        mnemonic
        for mnemonic, value in mnemonic_map.items()
        if (dqflags & value)
    }
    return mnemonics



import os

import asdf
import pytest
from astropy.time import Time


from roman_datamodels.testing.utils import mk_level2_image
from roman_datamodels.datamodels import ImageModel, FlatRefModel
from romancal.stpipe import RomanPipeline, RomanStep


@pytest.mark.parametrize("step_class", [RomanPipeline, RomanStep])
def test_open_model(step_class, tmp_path):
    """
    Test that the class is properly hooked up to datamodels.open.
    More comprehensive tests can be found in romancal.datamodels.tests,
    this is just a smoke test of the integration.
    """
    file_path = tmp_path / "test.asdf"

    with asdf.AsdfFile() as af:
        imod = mk_level2_image(arrays=(20, 20))
        af.tree = {'roman': imod}
        af.write_to(file_path)

    step = step_class()
    with step.open_model(file_path) as model:
        assert model.meta.telescope == "ROMAN"


@pytest.mark.skipif(
    os.environ.get("CI") == "true",
    reason="Roman CRDS servers are not currently available outside the internal network"
)
@pytest.mark.parametrize("step_class", [RomanPipeline, RomanStep])
def test_get_reference_file(step_class):
    """
    Test that CRDS is properly integrated.
    """
    im = mk_level2_image(arrays=(20, 20))
    # This will be brittle while we're using the dev server.
    # If this test starts failing mysteriously, check the
    # metadata values against the flat rmap.
    im.meta.instrument.optical_element = "F158"
    im.meta.observation.start_time = Time('2021-01-01T12:00:00')
    model = ImageModel(im)

    step = step_class()
    reference_path = step.get_reference_file(model, "flat")

    with step.open_model(reference_path) as reference_model:
        assert isinstance(reference_model, FlatRefModel)



"""
Entry point implementations.
"""


def get_steps():
    """
    Return tuples describing the stpipe.Step subclasses provided
    by this package.  This method is registered with the stpipe.steps
    entry point.

    Returns
    -------
    list of tuple (str, str, bool)
        The first element each tuple is a fully-qualified Step
        subclass name.  The second element is an optional class
        alias.  The third element indicates that the class
        is a subclass of Pipeline.
    """
    # Unit tests ensure that this list is kept in sync with the actual
    # class definitions.  We need to avoid importing romancal.pipeline and
    # romancal.step to keep the CLI snappy.
    return [
        ("romancal.step.DarkCurrentStep", None, False),
        ("romancal.step.FlatFieldStep", None, False),
    ]



"""
This module collects all of the stpipe.Step subclasses
made available by this package.
"""
from .dark_current.dark_current_step import DarkCurrentStep
from .flatfield.flat_field_step import FlatFieldStep


__all__ = [
    "DarkCurrentStep",
    "FlatFieldStep",
]



from os.path import basename
from setuptools import setup, find_packages
from glob import glob


NAME = 'romancal'

SCRIPTS = [s for s in glob('scripts/*') if basename(s) != '__pycache__']

PACKAGE_DATA = {
    '': [
        '*.fits',
        '*.txt',
        '*.inc',
        '*.cfg',
        '*.csv',
        '*.yaml',
        '*.json',
        '*.asdf'
    ]
}

setup(
    use_scm_version=True,
    setup_requires=['setuptools_scm'],
    scripts=SCRIPTS,
    packages=find_packages(),
    package_data=PACKAGE_DATA,
)



"""Project default for pytest"""
import os
import tempfile
import pytest
import inspect

from stdatamodels import s3_utils

#from jwst.associations import (AssociationRegistry, AssociationPool)
#from jwst.associations.tests.helpers import t_path
from romancal.lib.tests import helpers as lib_helpers


# @pytest.fixture(scope='session')
# def full_pool_rules(request):
#     """Setup to use the full example pool and registry"""
#     pool_fname = t_path('data/mega_pool.csv')
#     pool = AssociationPool.read(pool_fname)
#     rules = AssociationRegistry()
#
#     return (pool, rules, pool_fname)


@pytest.fixture
def mk_tmp_dirs():
    """Create a set of temporary directorys and change to one of them."""
    tmp_current_path = tempfile.mkdtemp()
    tmp_data_path = tempfile.mkdtemp()
    tmp_config_path = tempfile.mkdtemp()

    old_path = os.getcwd()
    try:
        os.chdir(tmp_current_path)
        yield (tmp_current_path, tmp_data_path, tmp_config_path)
    finally:
        os.chdir(old_path)


@pytest.fixture(autouse=True)
def monkey_patch_s3_client(monkeypatch, request):
    # If tmpdir is used in the test, then it is providing the file.  Map to it.
    if "s3_root_dir" in request.fixturenames:
        path = request.getfixturevalue("s3_root_dir")
    else:
        path = None
    monkeypatch.setattr(s3_utils, "_CLIENT", lib_helpers.MockS3Client(path))


@pytest.fixture
def s3_root_dir(tmpdir):
    return tmpdir


@pytest.fixture
def slow(request):
    """Setup slow fixture for tests to identify if --slow
    has been specified
    """
    return request.config.getoption('--slow')


@pytest.fixture(scope="module")
def jail(request, tmpdir_factory):
    """Run test in a pristine temporary working directory, scoped to module.

    This fixture is the same as _jail in ci_watson, but scoped to module
    instead of function.  This allows a fixture using it to produce files in a
    temporary directory, and then have the tests access them.
    """
    old_dir = os.getcwd()
    path = request.module.__name__.split('.')[-1]
    if request._parent_request.fixturename is not None:
        path = path + "_" + request._parent_request.fixturename
    newpath = tmpdir_factory.mktemp(path)
    os.chdir(str(newpath))
    yield newpath
    os.chdir(old_dir)


@pytest.mark.trylast
def pytest_configure(config):
    terminal_reporter = config.pluginmanager.getplugin('terminalreporter')
    config.pluginmanager.register(TestDescriptionPlugin(terminal_reporter), 'testdescription')


class TestDescriptionPlugin:
    """Pytest plugin to print the test docstring when `pytest -vv` is used.

    This plug-in was added to support JWST instrument team testing and
    reporting for the JWST calibration pipeline.
    """

    def __init__(self, terminal_reporter):
        self.terminal_reporter = terminal_reporter
        self.desc = None

    def pytest_runtest_protocol(self, item):
        try:
            # Get the docstring for the test
            self.desc = inspect.getdoc(item.obj)
        except AttributeError:
            self.desc = None

    @pytest.hookimpl(hookwrapper=True, tryfirst=True)
    def pytest_runtest_logstart(self, nodeid, location):
        # When run as `pytest` or `pytest -v`, no change in behavior
        if self.terminal_reporter.verbosity <= 1:
            yield
        # When run as `pytest -vv`, `pytest -vvv`, etc, print the test docstring
        else:
            self.terminal_reporter.write('\n')
            yield
            if self.desc:
                self.terminal_reporter.write(f'\n{self.desc} ')



import os

import pytest
import numpy as np

from roman_datamodels import stnode
from roman_datamodels.datamodels import ImageModel, FlatRefModel
from roman_datamodels.testing import utils as testutil
from romancal.flatfield import FlatFieldStep


@pytest.mark.parametrize(
    "instrument, exptype",
    [
        ("WFI", "WFI_IMAGE"),
    ]
)
@pytest.mark.skipif(
    os.environ.get("CI") == "true",
    reason="Roman CRDS servers are not currently available outside the internal network"
)
def test_flatfield_step_interface(instrument, exptype):
    """Test that the basic inferface works for data requiring a FLAT reffile"""

    shape = (20, 20)

    wfi_image = testutil.mk_level2_image(arrays=True)
    wfi_image.meta.instrument.name = instrument
    wfi_image.meta.instrument.detector = 'WFI01'
    wfi_image.meta.instrument.optical_element = 'F158'
    wfi_image.meta.exposure.type = exptype
    wfi_image.data = np.ones(shape, dtype=np.float32)
    wfi_image.dq = np.zeros(shape, dtype=np.uint32)
    wfi_image.err = np.zeros(shape, dtype=np.float32)
    wfi_image.var_poisson = np.zeros(shape, dtype=np.float32)
    wfi_image.var_rnoise = np.zeros(shape, dtype=np.float32)
    wfi_image.var_flat = np.zeros(shape, dtype=np.float32)
    wfi_image.area = np.ones(shape, dtype=np.float32)
    wfi_image_model = ImageModel(wfi_image)
    flatref = stnode.FlatRef()
    meta = {}
    testutil.add_ref_common(meta)
    meta['instrument']['optical_element'] = 'F158'
    meta['instrument']['dectector'] = 'WFI01'
    meta['reftype'] = 'FLAT'
    flatref['meta'] = meta
    flatref['data'] = np.ones(shape, dtype=np.float32)
    flatref['dq'] = np.zeros(shape, dtype=np.uint16)
    flatref['err'] = (np.random.random(shape) * 0.05).astype(np.float32)
    flatref_model = FlatRefModel(flatref)

    result = FlatFieldStep.call(wfi_image_model, override_flat=flatref_model)

    assert (result.data == wfi_image.data).all()
    assert result.var_flat.shape == shape
    assert result.meta.cal_step.flat_field == 'COMPLETE'



"""
Flat-field a science image
"""

from ..stpipe import RomanStep
from . import flat_field
import roman_datamodels as rdm


__all__ = ["FlatFieldStep"]


class FlatFieldStep(RomanStep):
    """Flat-field a science image using a flatfield reference image.
    """

    reference_file_types = ["flat"]

    def process(self, input):

        input_model = self.open_model(input)
        # Get reference file paths
        reference_file_names = {}
        reffile = self.get_reference_file(input_model, "flat")
        reference_file_names['flat'] = reffile if reffile != 'N/A' else None

        # Open the relevant reference files as datamodels
        reference_file_models = {}

        if reffile is not None:
            reference_file_models['flat'] = rdm.open(reffile)
            self.log.debug(f'Using FLAT ref file: {reffile}')
        else:
            reference_file_models['flat'] = None
            self.log.debug('Using FLAT ref file')

        # Do the flat-field correction
        output_model = flat_field.do_correction(
            input_model,
            **reference_file_models,
        )

        # Close the input and reference files
        input_model.close()
        try:
            for model in reference_file_models.values():
                model.close()
        except AttributeError:
            pass

        return output_model



from .referencefile import ReferenceFileModel
from ..dynamicdq import dynamic_mask


class FlatModel(ReferenceFileModel):
    """
    A data model for 2D flat-field images.

    Parameters
    __________
    data : numpy float32 array
         The science data

    dq : numpy uint16 array
         Data quality array

    err : numpy float32 array
         Error array
    """
    schema_url = "http://stsci.edu/schemas/roman_datamodel/reference_files/flat.schema"

    def __init__(self, init=None, **kwargs):
        super().__init__(init=init, **kwargs)

        # Saved for DQ code implementation
        self.dq = dynamic_mask(self)

        # Implicitly create arrays
        self.err = self.err
